#!C:\ci\theano_1528428757816\_h_env\python.exe
# EASY-INSTALL-ENTRY-SCRIPT: 'Theano==1.0.2+2.gc449c8699','console_scripts','theano-nose'
__requires__ = 'Theano==1.0.2+2.gc449c8699'
import re
import sys
from pkg_resources import load_entry_point

if __name__ == '__main__':
    sys.argv[0] = re.sub(r'(-script\.pyw?|\.exe)?$', '', sys.argv[0])
    sys.exit(
        load_entry_point('Theano==1.0.2+2.gc449c8699', 'console_scripts', 'theano-nose')()
    )
